#define MODULE "panic"

#include "panic/Panic.h"
#include "fs/InitRD.h"
#include "interrupts/IDT.h"
#include "io/PIC.h"
#include "log/Log.h"
#include "memory/VMM.h"
#include "misc/MSR.h"
#include "render/Framebuffer.h"
#include "render/TextRenderer.h"
#include "std/stdio.h"
#include "thread/Scheduler.h"
#include "trace/StackTracer.h"

static bool g_is_in_panic = false;
static bool g_is_in_double_panic = false;

void panic_prepare_keyboard_triple_fault()
{
    PIC::enable_master(0b11111101); // enable keyboard only
    PIC::enable_slave(0b11111111);

    IDTR idtr;
    idtr.limit = 0x0000;
    idtr.offset = 0x0000;
    asm volatile(
        "lidt %0"
        :
        : "m"(
            idtr)); // when an interrupt arrives, triple-fault (the only interrupt that should come is the keyboard one)

    asm volatile("sti");
}

void dump_registers(Context* context)
{
    kinfoln("-- Registers:");
    kinfoln("rax: %lx, rbx: %lx, rcx: %lx, rdx: %lx", context->rax, context->rbx, context->rcx, context->rdx);
    kinfoln("rsi: %lx, rdi: %lx, rsp: %lx, rbp: %lx", context->rsi, context->rdi, context->rsp, context->rbp);
    kinfoln("r8: %lx, r9: %lx, r10: %lx, r11: %lx", context->r8, context->r9, context->r10, context->r11);
    kinfoln("r12: %lx, r13: %lx, r14: %lx, r15: %lx", context->r12, context->r13, context->r14, context->r15);
    kinfoln("rip: %lx, cs: %lx, ss: %lx", context->rip, context->cs, context->ss);
    kinfoln("rflags: %lx, cr2: %lx", context->rflags, context->cr2);
    kinfoln("ia32_efer: %lx", MSR::read_from(IA32_EFER_MSR));
}

void fatal_dump_registers(Context* context)
{
    printf("-- Possibly Relevant Registers:\n");
    printf("rbp: %lx, rsp: %lx, rip: %lx, cs: %lx, ss: %lx, rflags: %lx, cr2: %lx, error code: %lx\n", context->rbp,
           context->rsp, context->rip, context->cs, context->ss, context->rflags, context->cr2, context->error_code);
}

[[noreturn]] void __panic_fatal_stub(Context* context)
{
    if (context) fatal_dump_registers(context);
    printf("-- No stack trace available\n");

    KernelLog::enable_log_backend(Backend::Console);
    KernelLog::toggle_log_backend(
        Backend::Console); // disable console logging, as we can page fault while writing to the framebuffer.

    Scheduler::current_task()->id = 0; // we're panicking, we don't even care anymore. This is so KernelLog shows
                                       // (kernel) instead of (taskname: PID) in the log.

    panic_prepare_keyboard_triple_fault();

    printf("Press any key to restart.\n");

    while (1) asm volatile("hlt");
}

[[noreturn]] void __panic_stub(Context* context)
{
    VMM::switch_back_to_kernel_address_space();
    VMM::apply_address_space();
    if (context) dump_registers(context);

    if (InitRD::is_initialized())
    {
        kinfoln("-- Stack trace:");
        if (context)
        {
            StackTracer tracer(context->rbp);
            tracer.trace_with_ip(context->rip);
        }
        else
        {
            uintptr_t rbp;

            asm volatile("mov %%rbp, %0" : "=r"(rbp));
            StackTracer tracer(rbp);
            tracer.trace();
        }
    }
    else { kinfoln("-- No stack trace available"); }

    panic_prepare_keyboard_triple_fault();

    while (1) asm volatile("hlt");
}

extern "C" [[noreturn]] bool __do_int_panic(Context* context, const char* file, int line, const char* message)
{
    asm volatile("cli");

    if (g_is_in_double_panic)
    {
        panic_prepare_keyboard_triple_fault();
        while (1) asm volatile("hlt");
    }

    if (g_is_in_panic)
    {
        g_is_in_double_panic = true;
        printf("Kernel panic while panicking (showing vital information only) at %s, line %d: %s\n", file, line,
               message);
        __panic_fatal_stub(context);
    }

    g_is_in_panic = true;

    KernelLog::enable_log_backend(Backend::Console);

    TextRenderer::reset();

    framebuffer0.clear(Color::Black);

    if (context->number >= 0x20 && context->number < 0x30) { PIC::send_eoi((uint8_t)(context->irq_number & 0xFF)); }

    Task* task;
    if ((task = Scheduler::current_task()))
    {
        kerrorln("Kernel panic in task %ld, at %s, line %d: %s", task->id, file, line, message);
    }
    else { kerrorln("Kernel panic at %s, line %d: %s", file, line, message); }

    __panic_stub(context);
}

extern "C" [[noreturn]] bool __do_panic(const char* file, int line, const char* message)
{
    asm volatile("cli");

    if (g_is_in_double_panic)
    {
        panic_prepare_keyboard_triple_fault();
        while (1) asm volatile("hlt");
    }

    if (g_is_in_panic)
    {
        g_is_in_double_panic = true;
        printf("Kernel panic while panicking (showing vital information only) at %s, line %d: %s\n", file, line,
               message);
        __panic_fatal_stub(nullptr);
    }

    g_is_in_panic = true;

    KernelLog::enable_log_backend(Backend::Console);

    TextRenderer::reset();

    framebuffer0.clear(Color::Black);

    Task* task;
    if ((task = Scheduler::current_task()))
    {
        kerrorln("Kernel panic in task %ld, at %s, line %d: %s", task->id, file, line, message);
    }
    else { kerrorln("Kernel panic at %s, line %d: %s", file, line, message); }

    __panic_stub(nullptr);
}